#  -------------------------------------------------------------------------------
#
#      Package PIKT::Random -- Utilities for generating Pseudo-Random Numbers
#
#  -------------------------------------------------------------------------------
#
#     This file is part of PIKT ("Problem Informant/Killer Tool").
#     Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.
#
#     rkey copyright (C) 2001,2002 by Michel Blanc.
#
#     PIKT is a trademark of the University of Chicago.  All Rights Reserved.
#
#     This program is free software; you can redistribute it and/or
#     modify it under the terms of the GNU General Public License
#     as published by the Free Software Foundation; either version 2
#     of the License, or (at your option) any later version.
#
#     PIKT is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program; if not, write to the Free Software
#     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
#     USA.
#
#     Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
#
#  -------------------------------------------------------------------------------
#
# $Id: Random.pm,v 1.5 2002/10/17 16:51:09 michel Exp $
#

package PIKT::Random;

use strict;
use POSIX;
use PIKT::Random::TT800;
use PIKT::Random::MD5;

@PIKT::Random::ISA = qw(Exporter);
@PIKT::Random::EXPORT = qw(psrand prand getscore);

my $VERSION;
my ( $ME, $ID, $DATE ) =  ( 'PIKT Random.pm', 
                            '$Id: Random.pm,v 1.5 2002/10/17 16:51:09 michel Exp $', 
                            '$Date: 2002/10/17 16:51:09 $' );

( $PIKT::Random::VERSION ) = '$Revision: 1.5 $ ' =~ /\$Revision:\s+([^\s]+)/;

my ($PRNG,$HASHER)=("") x 2;

push @INC, '.';

eval { require Math::Random::MT; };
if ($@) {
    $@ = undef;
    debug ('Using TT800 built-il rand as PRNG');
    $PRNG = 'TT800';
} else {
    debug ('using Mersenne Twister as PRNG');
    require Math::Random::MT;
    $PRNG = 'MT';   
}

$| = 1;

my $POOLSIZE = 220;
my $GENERATOR;

sub getscore () {

    my $verbose = shift || 0;
    my $score;
    my @scores = ("insecure - don't use it",      # 0
                  "probably insecure",            # 1
                  "possibly insecure",            # 2
                  "probably secure",              # 3
                  "secure",                       # 4
                  "very secure",                  # 5
                  "highly secure - best rating"); # 6

    my $verbiage = "";

    print "\nGuessing generated data quality rating :\n" if $verbose;
    print "\n\tLegend :" if $verbose;
    print "[.] : no effect,\n\t\t[+] : increases scoring,\n\t\t[-] : would increase scoring if fixed\n\n" if $verbose;

    if ( $PRNG eq "TT800" ) {
        $verbiage .= "\t[-] Math::Random::MT\n";
        $verbiage .= "\t[+] TT800\n";
        $score = 1;
    } elsif ( $PRNG eq "MT" ) {
        $verbiage .= "\t[+] Math::Random::MT\n";
        $score = 1;
    }
    if (-r '/dev/urandom') {
        $verbiage .= "\t[+] /dev/urandom\n";
        $score+=2;
    }
  
    my @res = psrand();

    if ( $res[0] > 9 ) { 
        $verbiage .= "\t[+] source succeeded count = $res[0]\n";
        $score++ ;
    } else {
        $verbiage .= "\t[-] source succeeded count < 10\n";
    }

    my $ssr =  int (($res[0] / ($res[1]+$res[0])) * 100);
    if ( $ssr > 79 ) { 
        $verbiage .= "\t[+] source success rate = $ssr% > 80%\n";
        $score++;
    } else {
        $verbiage .= "\t[-] source success rate = $ssr% < 80%\n";
    }

    print $verbiage if $verbose;
    print "\n\t$score";
    print "--",$scores[$score] if ($verbose);
    print "\n\t(note : this is just a rough informal figure)" if $verbose;
    print "\n\n";
}

sub psrand () {
    # Creating a core file
    # Generated an immediate core-dump.
    # Since we just forked, the child cores (fork return value is 0 if we're in child) and the father goes on.
    # This may not work, depending on ulimits
    fork or dump;
    
    my (@seeds, @sourcelist, %sources);
    my ($total_bytes, $failed_sources, $ok_sources, $srcline, $pool) = (0, 0, 0, '', '');
    
    %sources = ( 'common' => [ "head -n 200 $ENV{PWD}/core",
#                              'wget -O - -T 4 http://lavarand.sgi.com/seed/current/block.html -Y off',
                               'dmesg',
                               'uptime',],

                 'freebsd' => [ 'head -c 10k /dev/urandom', 
                                'netstat -m',
                                'ps -eaxlw ',
                                'netstat -in',
                                'pstat -t',
                                ],

                 'openbsd' => [ 'head -c 10k /dev/urandom', 
                                'netstat -m',
                                'ps -eaxlw ',
                                'netstat -in',
                                'pstat -t',
                                ],                 

                 'solaris' => [ 'ps -ef',
                                'ifconfig -a',
                                'netstat -n', 
                                'netstat -s',
                                'cat /usr/adm/messages',
                                'ps -eafl',
                                'netstat -in',
                                ],

                 'linux' => [ 'head -c 10k /dev/urandom', 
                              'arp -na',
                              'ps -ef',
                              'cat /var/log/messages',
                              'ps -eafl',
                              'head -c 10k /dev/urandom',
                              'lsof',
                              'vmstat -n',
                              'cat /proc/net/*',
                              'cat /proc/meminfo',
                              'cat /proc/uptime',
                              'cat /proc/slabinfo',
                              'ifconfig -a',
                              'netstat -n', 
                              'netstat -s',] , );
    
    #'nfsstat', 'netstat -m', 'netstat -in', 'mpstat', 'w',
    #'df', 'portstat', 'iostat', 'uptime', 'vmstat -f', 'vmstat',
    #'netstat -n', 'lsof', 'ps -el', 'ps aux', 'ipcs -a',
    #'pstat -p', 'pstat -S', 'pstat -v', 'pstat -x', 'pstat -t',
    #'last -n 50', 'last -50', 'history', 'arp -a', 'lpstat -t',
    #'wget -O - -T 4 http://lavarand.sgi.com/seed/current/block.html -Y off', 

    if (defined $sources{$^O}) {
        debug ("Using $^O source list");
        @sourcelist = (@{$sources{$^O}}, @{$sources{'common'}});
    } else {
        debug ("Unknown system $^O, using cumulative source list");
        foreach my $key (keys (%sources)) {
            push (@sourcelist, @{$sources{$key}});
        }
    }
    
#    print "Getting entropy....";
    
    foreach my $source (@sourcelist) {
        # If source has space (obviously, it contains a command ( [\ ] maniacs out :) )
        # or -x source is true, that's a command, else it is a single file
      ##  if (($source =~ /\S+\s+\S+/) or (-x "$source")) {
            unless (open (SRC,"$source 2>/dev/null |")) {
                debug ("Source '$source' failed !");
                $failed_sources++;
                next;
            }
        ##} else { # single file
        ##    unless (open (SRC,"<$source")) {
        ##        debug ("Source '$source' file failed !");
        ##        $failed_sources++;
        ##        next;
        ##    }
        ##}
        
        if (eof(SRC)) {
            debug ("Source '$source' failed !");
            $failed_sources++;  
            next;
        }
        
        $ok_sources++;
        debug ("Source '$source' succeeded");
        
        while (my $i=read (SRC, $srcline, $POOLSIZE) ) {
            # we pad shorter lines to poolsize
            while (length($srcline) < $POOLSIZE) { $srcline .= $srcline; }
            $srcline = substr($srcline, 0, $POOLSIZE);
            $pool ^= $srcline;
            $total_bytes += $i;
        };
        
        close(SRC);
    }
    
    debug ("core removal successful") if unlink 'core';
    
    for (my $i = 16; $i <= $POOLSIZE; $i += 32) {
        push(@seeds, unpack("L4", Digest(substr($pool, $i-16, $i+64))));
    }
    
    debug ("$total_bytes bytes read");
    debug ("$ok_sources sources succeeded, $failed_sources failed");

   # srand( time ^ $$ ^ unpack "%L*", `ps -e | gzip`);

    $GENERATOR = Math::Random::TT800->new ( @seeds[0 .. 24] ) if ($PRNG eq 'TT800');   
    $GENERATOR = Math::Random::MT->new ( unpack "L*", $seeds[0] ) if ($PRNG eq 'MT');
    srand( unpack "L*", $seeds[0] ) if ($PRNG eq 'perl');

    return ($ok_sources,$failed_sources);
}
               

sub prand () {
    my $factor = shift or 1;
    
    return $GENERATOR->next() * $factor if ($PRNG eq 'TT800');
    return $GENERATOR->rand($factor) if ($PRNG eq 'MT');
    return rand($factor) if ($PRNG eq 'perl');
}

sub debug {
    my ($message, $pid) = @_;
    my ($line, $caller) = ((caller(0))[2],(caller(1))[3]);
    my $pidstate;
    
    chomp ($message);
    
    return 0 unless $::DEBUG;
    
    if ($pid) {
        if (kill 0 => $pid) {
            $pidstate = "alive";
        } elsif ($! == EPERM) {
            $pidstate = "escaped [EPERM]";
        } elsif ($! == ESRCH) {
            $pidstate = "dead [ESRCH]";
        } else {
            $pidstate = "in unknown state - kill 0 failed !!!";
        }
    }
    
    my $when = time()-$^T;
    print STDERR "=DEBUG [$when secs]=>In $caller($line) : $message";
    print STDERR " (PID $pid is $pidstate)" if $pid;
    print STDERR "\n";
    
    return 1;
}
1;

__END__

=head1 NAME
 
PIKT::Random - PRNG routines for rkey & PIKT
 
=head1 SYNOPSIS
 
 use PIKT::Random;

 PIKT::Random::prand();
 $score = PIKT::Random::getscore();
 $randomnumber = PIKT::Random::prand($maxnumber);
 
=head1 FUNCTIONS

=item B<psrand> ()

This is the PIKT::Random seeder. Must be called before any call to prand.

=item B<prand> ([ARG])

Generates a random decimal number in the interval [0,ARG). If ARG is omitted, the interval is [0,1).

=item B<score> ()

Scores the random generation environment.
Currently defined scores are :

 0 -- insecure - don't use it
 1 -- probably insecure
 2 -- possibly insecure
 3 -- probably secure
 4 -- secure
 5 -- very secure
 6 -- highly secure - best rating

The score is derived from empiric computations, and should not be blindly trusted. For instance, it doesn't take into account the real entropy bit count generated by the pseudo random number sources.

=head1 DESCRIPTION
 
B<This module> provides methods to generate pseudo-random numbers (PRNs). Its goal is to provide stronger random routines than perl standard ones.

The modules' package also installs Math::Random::TT800 and PIKT::Random::MD5

=head1 ALGORITHM

To generate pseudo-random numbers PIKT::Random first try for find entropy sources (in psrand).
Actually, it has specific entropy sources for the following platforms :

 FreeBSD
 OpenBSD
 GNU/Linux
 Solaris

When psrand is run on another platform, B<ALL> entropy sources will be used. This should result in lower ranking in B<getscore>().

The resulting output of entropy sources are collected in an entropy pool, which is then hashed.
The different backend PRNGs will then be seeded from that pool. Currently supported PRNG backends are Math::Random::TT800 and Math::Random::MT.

Upon prand call, and PRN is generated using Math::Random::MT if it is available, or Math::Random::TT800 if not.
The resulting PRN will consitute I<half> of the final PRN. The second half is generated via std perl rand (see L<perlfunc(1)>), seeded by Larry Wall's method.

The goal is to have a final result generated by two independent RNG, independantly seeded, in order to reduce sequence guessing attacks.

=head1 CAVEATS
 
I am no crypto expert. I am no math expert. Those generated numbers *should* be unguessable. But you never know.
Beware of long-long hosts list. The longer the list, the bigger the exposure to sequence guessing.

 
=head1 BUGS

None known yet.

=head1 SEE ALSO

L<rkey(1)>, B<Math::Random::TT800>, B<PIKT HomePage> (http://www.pikt.org/)

=head1 AUTHORS
 
Michel Blanc <mblanc@erasme.org> for PIKT::Random

M. Matsumoto <matumoto@math.keio.ac.jp> for original TT800 C implementation (ftp://random.mat.sbg.ac.at/pub/data/tt800.c)

Otmar Lendl <lendl@cosy.sbg.ac.at> for TT800 translation to perl and enhancements to support multiple streams
of pseudorandom numbers

Alex Rousskov <rousskov@plains.NoDak.edu> for the pure-perl MD5 implementation

=head1 ACKNOWLEDGEMENTS
 
 Robert Osterlund as always, for deep testing
 Nick Russo for suggestions in <pikt-workers@gsb.uchicago.edu>
 
=cut
