/*
   mem.c -- PIKT memory routines

   $Header: /cvs/PIKT/pikt/src/lib/mem.c,v 1.11 2002/08/23 16:46:45 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "common.h"

/* malloc() the specified amount of memory; if unsuccessful, presumably
   because of virtual memory exhaustion, log a warning and keep trying;
   if successful, return a generic pointer to the newly allocated memory
 */
void *
mymalloc (size_t size, const char *func)
{
	static char fnnam[] = "mymalloc";
	void *ptr;
	static char msg[128];	/* should be plenty */
	char *msgoffset;

	while (! (ptr = malloc(size))) {
		if (isverboselog) {
			if (logdir && piktc_svclog) {	/* piktc_svclog is the
			                                   last path set in
			                                   initpaths()
			                                 */
				logreturn(PIKTLOG, LOG_WARNING, ERRID_MEMORY_ALLOCATION_FAILURE, __FILE__, __LINE__, fnnam, FALSE, "in %s(), malloc() failure", func);
			} else {
				exstat = maxval(exstat, exit_status[LOG_WARNING]);
				sprintf(msg, "%s %s %s[%d]: [ID %d %s:] (%s, line %d, %s()) in %s(), malloc() failure", get_timenow(), hostname, prgnam, (int)prgpid, ERRID_MEMORY_ALLOCATION_FAILURE, error_level[LOG_WARNING], __FILE__, __LINE__, fnnam, func);
				msgoffset = strchr(msg, ' ') + 1;
				fprintf(stderr, msgoffset);
				fflush(stderr);
			}
		}
		sleep(MALLOC_TIME_TO_SLEEP);
	}
	return ptr;
}

/* calloc() the specified amount of memory; if unsuccessful, presumably
   because of virtual memory exhaustion, log a warning and keep trying;
   if successful, return a generic pointer to the newly allocated memory
 */
void *
mycalloc (size_t nobj, size_t size, const char *func)
{
	static char fnnam[] = "mycalloc";
	void *ptr;
	static char msg[128];	/* should be plenty */
	char *msgoffset;

	while (! (ptr = calloc(nobj, size))) {
		if (isverboselog) {
			if (logdir && piktc_svclog) {	/* piktc_svclog is the
			                                   last path set in
			                                   initpaths()
			                                 */
				logreturn(PIKTLOG, LOG_WARNING, ERRID_MEMORY_ALLOCATION_FAILURE, __FILE__, __LINE__, fnnam, FALSE, "in %s(), calloc() failure", func);
			} else {
				exstat = maxval(exstat, exit_status[LOG_WARNING]);
				sprintf(msg, "%s %s %s[%d]: [ID %d %s:] (%s, line %d, %s()) in %s(), calloc() failure", get_timenow(), hostname, prgnam, (int)prgpid, ERRID_MEMORY_ALLOCATION_FAILURE, error_level[LOG_WARNING], __FILE__, __LINE__, fnnam, func);
				msgoffset = strchr(msg, ' ') + 1;
				fprintf(stderr, msgoffset);
				fflush(stderr);
			}
		}
		sleep(MALLOC_TIME_TO_SLEEP);
	}
	return ptr;
}

/* strdup() the string; if unsuccessful, presumably because of virtual memory
   exhaustion, log a warning and keep trying; if successful, return a char
   pointer to the newly allocated memory
 */
signed char *
mystrdup (const signed char *str, const char *func)
{
	static char fnnam[] = "mystrdup";
	static signed char *ptr;
	static char msg[128];	/* should be plenty */
	char *msgoffset;

	while (! (ptr = (signed char *) strdup(str))) {
		if (isverboselog) {
			if (logdir && piktc_svclog) {	/* piktc_svclog is the
			                                   last path set in
			                                   initpaths()
			                                 */
				logreturn(PIKTLOG, LOG_WARNING, ERRID_MEMORY_ALLOCATION_FAILURE, __FILE__, __LINE__, fnnam, FALSE, "in %s(), strdup() failure", func);
			} else {
				exstat = maxval(exstat, exit_status[LOG_WARNING]);
				sprintf(msg, "%s %s %s[%d]: [ID %d %s:] (%s, line %d, %s()) in %s(), strdup() failure", get_timenow(), hostname, prgnam, (int)prgpid, ERRID_MEMORY_ALLOCATION_FAILURE, error_level[LOG_WARNING], __FILE__, __LINE__, fnnam, func);
				msgoffset = strchr(msg, ' ') + 1;
				fprintf(stderr, msgoffset);
				fflush(stderr);
			}
		}
		sleep(MALLOC_TIME_TO_SLEEP);
	}
	return ptr;
}
