/*
   log.c -- PIKT pikt & piktd log routines

   $Header: /cvs/PIKT/pikt/src/piktd/log.c,v 1.33 2002/10/12 18:13:34 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"
#include "piktd.h"

const char *leveltypes[] = { "EMERG", "ALERT", "CRIT", "ERR", "WARNING", "NOTICE", "INFO", "DEBUG" };

int syslog_facility = LOG_LOCAL0;

static void logit(const int, const int, const int, const char *, const int, const char *, const int, const char *, va_list ap);

/* nonfatal error; log the message, then return to the caller
 */
void
logreturn (const int logfiletype, const int logtype, const int errid, const char *srcfile, const int lno, const char *fnnam, const int syscall, const char *fmt, ...)
{
	va_list ap;
	int dologit;

	exstat = maxval(exstat, exit_status[logtype]);
	dologit = TRUE;
	if (logfiletype != OUTPUT) {
		#ifdef PIKTD
		if (/* ! logdir || */ ! piktdlog) {
		#else
		if (/* ! logdir || */ ! piktlog) {
		#endif
			if ((logfiletype != SYSLOG) || ! isautosyslog) {
				dologit = FALSE;
			}
		}
	}
	if (dologit) {
		va_start(ap, fmt);
		logit(logfiletype, logtype, errid, srcfile, lno, fnnam, syscall, fmt, ap);
		va_end(ap);
	}
	return;
}

/* fatal error; log the message, then exit the program
 */
void
logexit (const int logfiletype, const int logtype, const int errid, const char *srcfile, const int lno, const char *fnnam, const int syscall, const char *fmt, ...)
{
	va_list ap;
	int dologit;

	exstat = maxval(exstat, exit_status[logtype]);
	dologit = TRUE;
	if (logfiletype != OUTPUT) {
		#ifdef PIKTD
		if (/* ! logdir || */ ! piktdlog) {
		#else
		if (/* ! logdir || */ ! piktlog) {
		#endif
			if ((logfiletype != SYSLOG) || ! isautosyslog) {
				dologit = FALSE;
			}
		}
	}
	if (dologit) {
		va_start(ap, fmt);
		logit(logfiletype, logtype, errid, srcfile, lno, fnnam, syscall, fmt, ap);
		va_end(ap);
	}
	#ifdef PIKTD
	rmlocks(FALSE, FALSE);
	#else
	rmlocks(interactive, FALSE);
	#endif
	exit(exstat);
}

/* format the message, adding a date/time stamp, the hostname, the
   program name, the program pid, an error id, a label, source file &
   source line number, if a serious error, and system error message,
   as appropriate; print the message, either to a file or to stderr
 */
static void
logit (const int logfiletype, const int logtype, const int errid, const char *srcfile, const int lno, const char *fnnam, const int syscall, const char *fmt, va_list ap)
{
	int errno_save;
	char buf[MAXBUF+1];	/* arbitrarily large */
	int buflen;
	char *bufoffset;
	FILE *logfile;

	errno_save = errno;
	fflush(stdout);
	strcpy(buf, get_timenow());
	buflen = (int)strlen(buf);
	bufoffset = strchr(buf, ' ') + 1;
	switch (logtype) {
		case LOG_EMERG:
		case LOG_ALERT:
		case LOG_CRIT:
		case LOG_ERR:
			sprintf(buf+buflen, " %s %s[%d]: [ID %d %s] (%s, line %d, %s()) ", hostname, prgnam, (int)prgpid, errid, error_level[(logtype != NIL) ? logtype : LOG_ERR], srcfile, lno, fnnam);
			break;
		case LOG_DEBUG:
			sprintf(buf+buflen, " %s %s[%d]: [ID %d %s] (%s, line %d, %s()) ", hostname, prgnam, (int)prgpid, errid, error_level[(logtype != NIL) ? logtype : LOG_DEBUG], srcfile, lno, fnnam);
			break;
		default:
			sprintf(buf+buflen, " %s %s[%d]: [ID %d %s] ", hostname, prgnam, (int)prgpid, errid, error_level[(logtype != NIL) ? logtype : LOG_DEBUG]);
			break;
	}
	vsprintf(buf+strlen(buf), fmt, ap);
	#if defined(SHOW_SYSCALL)
	if (syscall)
		sprintf(buf+strlen(buf), ": %s", strerror(errno_save));
	#endif
	strcat(buf, "\n");
	requote(buf);
	switch (logfiletype) {
		case OUTPUT:
			fputs(bufoffset, stderr);
			fflush(stderr);
			break;
		case PIKTLOG:
			if (ischeck || isdoublecheck) {
				fputs(bufoffset, stderr);
				fflush(stderr);
			} else {
				#ifdef PIKTD
				if ((logfile = fopen(piktdlog, "a")) == NULL) {
					logexit(OUTPUT, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, "logit", TRUE, "fopen() error for %s", piktdlog);
				}
				fputs(bufoffset, logfile);
				if (fclose(logfile)) {
					logexit(OUTPUT, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, "logit", TRUE, "fclose() error for %s", piktdlog);
				}
				chmod(piktdlog, (mode_t)(S_IRUSR|S_IWUSR));  /* 600 */
				#else
				if (isdebug) {
					fputs(bufoffset, stderr);
					fflush(stderr);
				} else {
					if ((logfile = fopen(piktlog, "a")) == NULL) {
						logexit(OUTPUT, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, "logit", TRUE, "fopen() error for %s", piktlog);
					}
					fputs(bufoffset, logfile);
					/* if interactive, send error
					   message also to stderr
					 */
					if (interactive && ((logtype == LOG_ERR) /* || (logtype == WARNING) */ )) {
						fputs(bufoffset, stderr);
						fflush(stderr);
					}
					if (fclose(logfile)) {
						logexit(OUTPUT, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, "logit", TRUE, "fclose() error for %s", piktlog);
					}
					chmod(piktlog, (mode_t)(S_IRUSR|S_IWUSR));  /* 600 */
				}
				#endif
			}
			break;
		#ifndef PIKTD
		case LOCLOG:
			if (isdebug) {
				fputs(bufoffset, stderr);
				fflush(stderr);
			} else {
				if ((logfile = fopen(stmcurr->logfile, "a")) == NULL) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, "logit", TRUE, "fopen() error for %s", stmcurr->logfile);
					return;
				}
				fputs(bufoffset, logfile);
				if (fclose(logfile)) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, "logit", TRUE, "fclose() error for %s", stmcurr->logfile);
					return;
				}
				chmod(stmcurr->logfile, (mode_t)(S_IRUSR|S_IWUSR)); /* 600 */
			}
			break;
		case SYSLOG:
			if (isdebug) {
				fputs(bufoffset, stderr);
				fflush(stderr);
			} else {
				if (! isautosyslog) {
					openlog("pikt", LOG_CONS|LOG_PID, syslog_facility);
					syslog(LEVEL, "%s", buf+buflen+1);
					closelog();
				}
			}
			break;
		default:
			break;
		#endif
	}
	if (isautosyslog) {
		openlog(prgnam, LOG_CONS|LOG_PID, syslog_facility);
		syslog((logtype != NIL) ? logtype : LOG_DEBUG, "%s", buf+20);
		closelog();
	}
	return;
}
